=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright 2008 Fredo6 - Designed and written December 2008 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:  Lib6Plugin.rb
# Original Date	:  10 Dec 2008 - version 3.0
# Type			:  Script library part of the LibFredo6 shared libraries
# Description	:  A utility library about Plugin Configuration for LibFredo6-compliant scripts.
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module Traductor

#--------------------------------------------------------------------------------------------------------------
# PLugin Management (Part II)
#--------------------------------------------------------------------------------------------------------------			 				   

T6[:T_HELP_MenuDoc] = "Documentation"	
T6[:T_HELP_License] = "License"	
T6[:T_HELP_TrialExpiry] = "Trial Period Expiration"	
T6[:T_HELP_MenuVideo] = "Video"	
T6[:T_HELP_MenuAbout] = "About"	
T6[:T_HELP_MenuLanguages] = "Set Preferred Languages"	
T6[:T_HELP_MenuDefParam] = "Default Parameters"	
T6[:T_HELP_MenuTraceLog] = "View Trace log files"	
T6[:T_TIP_MenuTraceLog] = "Consult the log of traces for the current and last sessions"	
T6[:T_HELP_MenuDebugLog] = "View Debug log files"	
T6[:T_TIP_MenuDebugLog] = "Consult the log of debug messages for the current and last sessions"	
T6[:T_HELP_MenuPurgeObsolete] = "Purge Obsolete files"	
T6[:T_TIP_MenuPurgeObsolete] = "Identify old files and allow to delete them"	
T6[:T_TIP_MenuLicense] = "Get a License or check the license status"	
T6[:T_HELP_MenuPerformances] = "Performances"	
T6[:T_TIP_MenuPerformances] = "Summarize the load times for the plugins managed by LibFredo6"	
T6[:T_HELP_MenuTranslation] = "Language Translation"	
T6[:T_HELP_MenuDonation] = "Donation"
T6[:T_HELP_MenuWebInfo] = "Plugin Information"

T6[:T_HELP_Info] = "INFORMATION"
T6[:T_HELP_RootPath] = "Plugin Root Directory"
T6[:T_HELP_ParamPath] = "Default Parameters Directory"
T6[:T_HELP_TempPath] = "Temporary Files Directory"
T6[:T_HELP_Subfolder] = "Plugin Subfolder"
T6[:T_HELP_PluginLoadTime] = "Time to Load Plugins"
T6[:T_HELP_LoadTime] = "Total time to load the Plugin"
T6[:T_HELP_TotalLoadTime] = "Total load time of ALL LibFredo6 registered plugins at SU startup"
T6[:T_HELP_Usage] = "Total usage"
T6[:T_HELP_PLUGIN_USE] = "PLUGIN USE"
T6[:T_HELP_LibFredo6] = "LibFredo6 version"
T6[:T_HELP_Designed] = "Designed and developed by %1 - %2"
T6[:T_HELP_CreatedBy] = "Created by %1"
T6[:T_HELP_Date] = "Date"
T6[:T_HELP_Credits] = "CREDITS"
T6[:T_HELP_Support] = "SUPPORT"
T6[:T_HELP_RepositoryName] = "Plugin Repository"		 
T6[:T_HELP_WebSiteName] = "Web Site"		 
T6[:T_HELP_WebSiteLink] = "Web Site Link"		 
T6[:T_HELP_WebSupportLink] = "Check for updates"		 

T6[:T_HELP_CheckOlder] = "Checking older versions of Plugin"
T6[:T_HELP_PluginFolder] = "Sketchup Plugin Folder:"
T6[:T_HELP_FileFound] = "The following files are older and can be removed"
T6[:T_HELP_DirFound] = "The following subfolders are older and can be removed"

T6[:T_HELP_LoadingModules] = "Loading all Ruby files for plugin %1"
T6[:T_HELP_LoadedModules] = "Loaded all Ruby files for plugin %1 --> %2"
T6[:T_HELP_StartupTime] = "Startup"

T6[:T_ERROR_VersionModule] = "%1: requires LibFredo6 version >= %2"
T6[:T_ERROR_VersionSketchup] = "%1: requires Sketchup version >= %2"

T6[:DONATION_1] = "%1 is a free plugin for private and commercial use." 
T6[:DONATION_2] = "So if you donate, I'll take it as a sign of reward and recognition."
T6[:DONATION_3] = "Open the Donation page [%1]"

T6[:T_MSG_SCFLicenseInstall] = "Please install Sketchucation ExtensionStore v3.1 or above to License %1"

T6[:MSG_TrialExpiryCheck] = "Checking trial expiry date"
	
T6[:MSG_TEM_PleaseUpgrade] = "To continue using the extension %1, please upgrade to the latest version of Sketchucation ExtensionStore"	

#=====================================================================================
#=====================================================================================
# CLASS PluginUtils
#=====================================================================================
#=====================================================================================

class PluginUtils

def initialize(plugin)
	@plugin = plugin
end

#Modify the name of the action call_back (default is "action_mapping".
#The method name must be passed as a string or a symbol
def set_action_callback(method_name)
	return unless method_name
	@action_call_back_name = method_name.to_s
	@action_call_back_proc = nil
end

#Launch an action for the plugin
#This calls the plugin call back for mapping actions	
def launch_action(*args)
	#Check the license
	if !args.include?(:selection_menu)
		#puts "Check License #{@plugin.plugin_name.inspect} - args = #{args} - plugin = #{@plugin_name}"
		rootname = @plugin.plugin_name
		ckey = rand(100)
		cstamp = @plugin.tem_launcher(ckey)
		stamp0 = Math.sqrt(ckey**1.5 + 2)
		#puts "LAUNCH xstamp = #{cstamp} - stamp0 = #{stamp0}"
		return if stamp0 != cstamp
	end
	
	#Loading the full module if not done already
	return unless @plugin.load_second_phase_rubies
	
	#Loading the common libraries
	Traductor.use_libraries
	
	#Finding the mapping call back
	unless @action_call_back_proc
		@action_call_back_name = "action__mapping" unless @action_call_back_name
		@action_call_back_proc = @plugin.get_module.method @action_call_back_name
		return unless @action_call_back_proc
	end
	
	#Loading the themes
	@plugin.theme_load

	#Invoking the action
	@action_call_back_proc.call *args
end
	
#Check the class is OK	
def PluginUtils.all_activated?(*args)
	args[1] * (Time.now.to_f.round - 5000)
end
	
#Checking License
def PluginUtils.configured?(rootname)
	#If SCF License is not installed, consider that it is NOT licensed
	return nil unless defined?(SCFLicense)
	
	#Check the version of SCFLicense
	#puts "SFC version = #{SCFLicense.visual_version.inspect} - num = #{SCFLicense.visual_version.to_f.inspect}"
	
	#Building a dynamic token
	srand
	n = rand(99999)
	dynamic_token = SCFLicense.dynamic_token "sBUSkh8QYICsKv0Mstpf0ShOM2Kgvk", n
	return nil unless dynamic_token
	
	#Calling SCFLicense to check if the plugin is licensed
	type, key = SCFLicense.licensed? rootname, dynamic_token
	#puts "PLUGIN LICENSED? Type = #{type} key = #{key}"
	(key == "Masque_#{n}") ? type : nil
end

def PluginUtils.request_config(rootname)
end

#Load all Body files which have not been loaded yet
def PluginUtils.load_all_body_files(folder, rootname)
	lf = []
	path = File.join(LibFredo6.sudir, folder, "body_*")
	LibFredo6.dir_of_rb_files(path).each do |f|
		basef = File.basename f
		lf.push File.join(folder, basef) unless $".find { |s| s.include?(basef) || s.downcase.include?(basef.downcase)} 
	end	
	lf.each do |f|
		LibFredo6.require_ruby f
	end
	nil	
end

end	#class PluginUtils
		
#==========================================================================
#--------------------------------------------------------------------------
# Class Plugin: Define and Manage a plugin for LibFredo6
#--------------------------------------------------------------------------
#==========================================================================
	
class Plugin

def get_module
	@hmod_main_module
end

def eval_T6(symb, *args)
	t6 = T6Mod.handler @main_module
	t6.get(symb, *args) if t6
end

def eval_constant(symb)
	hmod = eval @main_module
	begin
		return ((hmod.class == Module) ? hmod.const_get(symb) : nil)
	rescue
		return nil
	end	
end

#--------------------------------------------------------------------------
# Menu and Action registration
#--------------------------------------------------------------------------

#Modify the name of the action call_back (default is "action_mapping".
#The method name must be passed as a string or a symbol
def set_action_callback(method_name)
	@putils = PluginUtils.new(self) unless @putils
	@putils.set_action_callback(method_name)
end
	
#Launch an action for the plugin
#This calls the plugin call back for mapping actions	
def launch_action(*args)
	Traductor.use_libraries
	n1 = rand(9)
	return if (n1 * (Time.now.to_f.round - 5000) - PluginUtils.all_activated?(rand(999), n1)).abs > 2 * n1
	@putils = PluginUtils.new(self) unless @putils
	@putils.launch_action(*args)
end
	
#--------------------------------------------------------------------------
# On the fly loading
#--------------------------------------------------------------------------

#Load a Body ruby file on the fly
def load_body(classe, file)
	return if @hsh_loaded_rubies[file]
	@hsh_loaded_rubies[file] = true
	basename = File.basename(file, ".*")
	body_name = "body_#{basename}"
	fname = File.join @folder, "body_#{basename}"
	t0 = Time.now
	LibFredo6.require_ruby fname
	delta = Time.now - t0
	add_load_time_second delta
	LibFredo6.log "#{basename}: Loading body code from #{body_name} [#{(delta * 1000).to_i} ms]"
end

#Load all Body files which have not been loaded yet
def load_all_body_files
	PluginUtils.load_all_body_files(@folder, @rootname)
end

#Return the Plugin folder
def get_plugin_dir
	@plugin_dir
end
	
def get_binaries_dir
	File.join(@plugin_dir, 'Binaries')
end
	
#--------------------------------------------------------------------------
# Ancillary menus for Plugins
#--------------------------------------------------------------------------

#Get the load time at SU starup for the plugin
def get_load_time
	text = "#{(@load_time * 1000.0).to_i} ms"
	text << " [#{T6[:T_HELP_StartupTime]}] + #{(@load_time_second * 1000.0).to_i} ms" if @load_time_second
	text
end

#Return load time info in ms
def load_time_info
	t = []
	t[0] = @load_time - @load_time_startup
	t[1] = @load_time_startup
	t[2] = t[0] + t[1]
	t[3] = (@load_time_second) ? @load_time_second : 0
	t[4] = t[2] + t[3]
	t = t.collect { |a| (a * 1000).to_i }
	t[3] = nil if t[3] == 0
	t
end

#Register additional time for Phase 2 loading
def add_load_time_second(t)
	@load_time_second = 0.001 unless @load_time_second
	@load_time_second += t
	AllPlugins.update @plugin_name, { :load_time_second => @load_time_second }
end

#Add a credit
def add_credits(credits)
	return unless credits
	@lst_credits = [] unless @lst_credits
	credits = [credits] unless credits.class == Array
	@lst_credits += credits
end

#Add a video
def add_videos(videos)
	return unless videos
	@lst_video = [] unless @lst_video
	videos = [videos] unless videos.class == Array
	@lst_video += videos
end

#Create the menu items	for all support utilities				
def populate_support_menu(cmdfamily, submenu=nil, plugin_title_symb=nil)
	@plugin_title_symb = plugin_title_symb if plugin_title_symb 

	#Specific to LibFredo6
	if @plugin_name =~ /LibFredo6/i
		text = T6[:T_HELP_MenuLanguages] + "..."
		cmdfamily.add_command(text, text, nil, nil, nil) { T6Mod.dialog_preferred_languages }
		
		if SU_MAJOR_VERSION >= 6
			cmdfamily.add_menu_separator
			text = T6[:T_BUTTON_CheckForUpdate] + "..."
			tooltip = @web_support_link
			hoptions = { :origin_plugin => @plugin_name }
			cmdfamily.add_command(text, tooltip, nil, nil, nil) { Upgrade.top_dialog hoptions }
			
			cmdfamily.add_menu_separator
			text = T6[:T_HELP_MenuPurgeObsolete] + "..."
			tooltip = T6[:T_TIP_MenuPurgeObsolete]
			cmdfamily.add_command(text, tooltip, nil, nil, nil) { PurgeObsoleteDialog.invoke }
			
			text = T6[:T_HELP_MenuPerformances] + "..."
			tooltip = T6[:T_TIP_MenuPerformances]
			cmdfamily.add_command(text, tooltip, nil, nil, nil) { PerformanceDialog.invoke }
			
			cmdfamily.add_menu_separator
			text = T6[:T_HELP_MenuTraceLog] + "..."
			tooltip = T6[:T_TIP_MenuTraceLog]
			cmdfamily.add_command(text, tooltip, nil, nil, nil) { TraceLogDialog.invoke }
			
			if MYDEFPARAM[:T_DEFAULT_DebugShowMenu]
				text = T6[:T_HELP_MenuDebugLog] + "..."
				tooltip = T6[:T_TIP_MenuDebugLog]
				cmdfamily.add_command(text, tooltip, nil, nil, nil) { DebugLogDialog.invoke }
			end	
		end	
		cmdfamily.add_menu_separator
	end
	
	#About dialog box
	cmdfamily.add_menu_separator
	text = "#{T6[:T_HELP_MenuAbout]} #{@name} v#{@version}..."
	cmdfamily.add_command(text, text, nil, nil, nil) { self.show_about }

	#Information for Sketchup support
	text = "Fredo6 Plugins Information for Sketchucation Support"
	cmdfamily.add_command(text, text, nil, nil, nil) { Fredo6.info }
	
	#SCF License
	if @scf_license
		text = "#{T6[:T_HELP_License]}..."
		cmdfamily.add_command(text, text, nil, nil, nil) { show_license }	
		cmdfamily.add_menu_separator if @lst_video.length > 0
	end
	
	#In Sketchup 5, this is the only thing we can do
	return unless SU_MAJOR_VERSION >= 6 
	
	#Documentation
	menu_documentation cmdfamily

	#Videos
	video_menu cmdfamily if @lst_video.length > 0
	
	#Web Site and Support Links
	cmdfamily.add_menu_separator if @web_site_link || @web_support_link
	
	if @web_repository_link 
		repository = (@web_repository) ? " [#{@web_repository}]" : ""
		text = T6[:T_HELP_RepositoryName] + repository + "..."
		tooltip = @web_repository_link
		cmdfamily.add_command(text, tooltip, nil, nil, nil) { open_URL @web_repository_link }
	end
	
	#Check for updates
	if @web_support_link && @web_support_link.strip != ""
		forum = (@web_site_name) ? " [#{@web_site_name}]" : ""
		text = T6[:T_HELP_MenuWebInfo] + forum + "..."
		tooltip = @web_support_link
		cmdfamily.add_command(text, tooltip, nil, nil, nil) { open_URL @web_support_link }
		unless @plugin_name =~ /LibFredo6/i
			text = T6[:T_BUTTON_CheckForUpdate] + "..."
			hoptions = { :origin_plugin => @plugin_name }
			cmdfamily.add_command(text, tooltip, nil, nil, nil) { Upgrade.top_dialog hoptions }
		end	
		
	elsif @web_site_link && @web_site_link.strip != ""
		forum = (@web_site_name) ? " [#{@web_site_name}]" : ""
		text = T6[:T_HELP_WebSiteName] + forum + "..."
		tooltip = @web_site_link
		cmdfamily.add_command(text, tooltip, nil, nil, nil) { open_URL @web_site_link }
	end
	
	#Donation
	if @donation
		donation_menu cmdfamily
	end
		
	#Default Parameters and Translation
	cmdfamily.add_menu_separator
	text = T6[:T_HELP_MenuDefParam] + "..."
	cmdfamily.add_command(text, text, nil, nil, nil) { invoke_default_parameter_dialog }

	text = T6[:T_HELP_MenuTranslation] + "..."
	cmdfamily.add_command(text, text, nil, nil, nil) do
		T6Mod.visual_edition @rootname #if load_second_phase_rubies
	end	
end

#Call the default parameter dialog box
def invoke_default_parameter_dialog
	return unless load_second_phase_rubies
	@defparam.visual_edition eval_T6(@plugin_title_symb), @plugin_name
end

#Manage menu entries for videos
def video_menu(cmdfamily)
	lsv = []
	@lst_video.each_with_index do |sv, i|
		next unless sv && sv.strip != ''
		titv = nil
		url = sv.strip
		if sv =~ /;/
			titv = $`.strip if $` 
			url = $'.strip
		end
		unless url =~ /\Ahttp\:\/\//i || url =~ /\Ahttps\:\/\//i
			url = $' if url =~ /file\:\/\//i
			vfile = File.join @plugin_dir, url
			next unless FileTest.exist? vfile
			url = "file://" + vfile 
		end
		titv = "Video #{i+1}" unless titv && titv != ''
		lsv.push [Traductor[titv], url]
	end
	if lsv.length > 0
		text = T6[:T_HELP_MenuVideo] + "..."
		menuvideo = cmdfamily.add_submenu(text)
		if menuvideo
			lsv.each do |l| 
				cmd = UI::Command.new(l[0]) { Traductor.openURL l[1] }
				cmd.status_bar_text = l[1]
				menuvideo.add_item cmd 			
			end	
		end	
	end	
end

#Manage menu entries for videos
def donation_menu(cmdfamily)
	lsv = []
	@lst_donation.each_with_index do |sv, i|
		next unless sv && sv.strip != ''
		next if sv =~ /\Afredo/i
		return if sv =~ /\Anone/i
		titv = nil
		url = sv.strip
		if sv =~ /;/
			titv = $`.strip if $` 
			url = $'.strip
		end
		unless url =~ /\Ahttp\:\/\//i || url =~ /\Ahttps\:\/\//i
			url = $' if url =~ /file\:\/\//i
			vfile = File.join @plugin_dir, url
			next unless FileTest.exist? vfile
			url = "file://" + vfile 
		end
		titv = "Donation #{i+1}" unless titv && titv != ''
		lsv.push [Traductor[titv], url]
	end
	lsv = [["Fredo6 at Paypal", @plugin_name]] if lsv.length == 0
	cmdfamily.add_menu_separator
	text = T6[:T_HELP_MenuDonation] + "..."
	menudonation = cmdfamily.add_submenu(text)
	if menudonation
		lsv.each do |l| 
			tit, url = l
			cmd = UI::Command.new(tit) { execute_donation(url) }
			cmd.status_bar_text = T6[:DONATION_3, url]
			menudonation.add_item cmd 			
		end	
	end	
end

#Invoke the donation page
def execute_donation(url)	
	return Traductor.openURL(url) if url =~ /:\/\//
	file = File.join LibFredo6.path, "Fredo6_donation.html"
	lines = IO.readlines(file)
	t = T6[:DONATION_1, url] + "<br>" + T6[:DONATION_2]
	text = ""
	lines.each { |s| text << s.sub("!text!", t) }
	tmpfile = File.join Traductor.temp_dir, "donation.html"
	File.open(tmpfile, "w") { |f| f.puts text }
	Traductor.open_doc tmpfile
end

#=================================================
# Documentation Management
#=================================================

#Compute the documentation for the Plugin
def compute_documentation
	@hsh_doc_info = {}
	
	#PDF Doc in the Plugin directory
	ldocs = []
	supath = File.join @plugin_dir, "*.pdf"
	ldocs += Dir[supath]
	supath = File.join @plugin_dir, "Documentation", "*.*"
	ldocs += Dir[supath]
	supath = File.join LibFredo6.sudir, "*#{@plugin_name}*.pdf"
	ldocs += Dir[supath]
	
	#Adding remote links if any
	ldocs += @lst_doclinks if @lst_doclinks
	
	#Building the hash key for the documentation
	ldocs.sort.each { |f| filter_doc_entry(f) }
end

#Register the documentation pattern for a sub plugin
def register_documentation_pattern(subplug, pattern)
	return unless pattern
	pattern = [pattern] if pattern.class != Array
	pattern.each do |pat|
		@lst_doc_pattern = {} unless @lst_doc_pattern
		pat = Regexp.new(pat, Regexp::IGNORECASE) if pat.class == String
		@lst_doc_pattern.push [subplug, pat]
	end	
end

def filter_doc_entry(file_or_url)
	#Parsing the specs for documentation file entry
	lf = file_or_url.split ";"
	lf = lf.collect { |s| s.strip }
	if lf.length == 1
		bf = nil
		f = lf[0]
	else
		bf, f = lf
	end	
	unless bf
		if FileTest.exist?(f)
			bf = File.basename(f)
		elsif f =~ /http/i
			bf = f
		else
			return 
		end
	end	

	#Assigning the right key
	@lst_doc_pattern.each do |a|
		key, pat = a
		return create_documentation_entry(key, bf, f) if bf =~ pat
	end
	pat = (@doc_pattern) ? @doc_pattern : @name.gsub(' ', '')
	pat = Regexp.new(pat, Regexp::IGNORECASE) if pat.class == String
	create_documentation_entry(nil, bf, f) if bf =~ pat	
end

#Create a documentation entry
def create_documentation_entry(key, bf, f)	
	lst = @hsh_doc_info[key]
	lst = @hsh_doc_info[key] = [] unless lst
	a = [key, bf, f]
	lst.push a unless lst.include?(a)
end

#Compute the menu for documentation
def menu_documentation(cmdfamily)
	compute_documentation
	return if @hsh_doc_info.empty?
	
	text = T6[:T_HELP_MenuDoc] + "..."
	menudoc = cmdfamily.add_submenu(text)
	
	@hsh_doc_info.each do |key, lst| 
		if key
			submenu = menudoc.add_submenu key
		else
			submenu = menudoc
		end	
		next unless submenu
		lst.each do |a|
			key, bf, f = a
			submenu.add_item(bf) { Traductor.openURL f }
		end	
	end	
end

#Get the list of documentation info for a given key
def get_documentation_info(key)
	@hsh_doc_info[key]
end

#Open a URL with complete and correct form
def open_URL(url)
	Traductor.openURL url
end

#=================================================
# THEME: Themes Management
#=================================================

def theme_list
	@thememan = ThemeManager.instantiate unless @thememan
	@thememan.theme_list_from_plugin(@plugin_name)
end

def theme_load(theme_name=nil)
	theme_name = @defparam[:T_PLUGIN_THEME] unless theme_name
	return unless theme_name
	@thememan = ThemeManager.instantiate unless @thememan
	@thememan.register_from_plugin(@plugin_name)
	@thememan.get_from_plugin(@plugin_name, theme_name)
end

#THEME: get a theme for the plugin. Default Theme is used if argument is nil
def theme_get(theme_name=nil)
	theme_name = @defparam[:T_PLUGIN_THEME] unless theme_name
	return {} unless theme_name
	return theme_load(theme_name) unless @thememan
	@thememan.get_from_plugin(@plugin_name, theme_name)
end

#=================================================
# Pictures Management
#=================================================

#Get Image directory list, based on preference
def picture_all_folders
	@picturefamily = Traductor::PictureFamily.new(@folder, @picture_prefix) unless @picturefamily
	@picturefamily.all_folders
end

def picture_selected_folders(selected_folders=nil)
	@picturefamily = Traductor::PictureFamily.new(@folder, @picture_prefix) unless @picturefamily
	@picturefamily.selected_folders selected_folders
end

def declare_picture_folders_symb(symb)
	@picture_symb = symb
end

#Get a picture: priority to SVG/PDF, then png, then _32, _24, _16 png
#               ...unless the extension is passed within the filename
def picture_get(filename)
	return nil unless filename
	filename = filename.strip
	return nil if filename == ""
	folders = nil
	folders = @defparam[@picture_symb] if @defparam && @picture_symb
	list_dir_pictures = picture_selected_folders folders
	
	#List of extensions in order
	if filename =~ /\.\w*\d*\Z/
		lst_ext = ['']
	else	
		#lst_ext = [(RUN_ON_MAC) ? '.pdf' : '.svg'] + ['.png']
		lst_ext = [(RUN_ON_MAC) ? '.svg' : '.svg'] + ['.png']
	end	
	
	#Checking filename existence
	lst_num = ['', '_32', '_24', '_16']
	list_dir_pictures.each do |d|
		lst_ext.each do |ext|
			if ext == '.png'
				lst_num.each do |num|
					imgfile = filename + num + ext 
					f = File.join @su_plugin_dir, d, imgfile
					return f if FileTest.exist?(f)
				end
			else	
				imgfile = filename + ext 
				f = File.join @su_plugin_dir, d, imgfile
				return f if FileTest.exist?(f)
			end	
		end	
	end

	#lst_ext.each do |ext|
	#	imgfile = filename + ext 
	#	list_dir_pictures.each do |d|
	#		f = File.join @su_plugin_dir, d, imgfile
	#		return f if FileTest.exist?(f)
	#	end
	#end	
	return nil if lst_ext[0] == ''
	
	#Trying with pixel number
	#ext = '.png'
	#lst_num = [32, 24, 16]
	#list_dir_pictures.each do |d|
	#	lst_num.each do |num|
	#		imgfile = filename + "_#{num}" + ext 
	#		f = File.join @su_plugin_dir, d, imgfile
	#		return f if FileTest.exist?(f)
	#	end	
	#end
	#lst_num.each do |num|
	#	imgfile = filename + "_#{num}" + ext 
	#	list_dir_pictures.each do |d|
	#		f = File.join @su_plugin_dir, d, imgfile
	#		return f if FileTest.exist?(f)
	#	end
	#end	

	return nil	
end

#Find a list of picture in the logical folders for the plugin
def picture_find_all(pattern, lst_ext=nil)
	return [] unless pattern
	pattern = pattern.strip
	return [] if pattern == ""
	
	#Directories where to find
	folders = nil
	folders = @defparam[@picture_symb] if @defparam && @picture_symb
	list_dir_pictures = picture_selected_folders folders
	#puts "list dir = #{list_dir_pictures}"
	
	#List of extensions in order
	ext_svg = (RUN_ON_MAC) ? 'pdf' : 'svg'
	unless lst_ext
		lst_ext = ['gif', ext_svg, 'png', 'jpg']
	end	
	if pattern =~ /\.\w*\d*\Z/
		lst_ext = ['']
	end	
	
	#Checking filename existence
	lst_all_files = []
	list_dir_pictures.each do |d|
		lst_ext.reverse.each do |ext|
			imgfile = pattern 
			imgfile += '.' + ext unless ext.empty?
			fdir = File.join @su_plugin_dir, d, imgfile
			lst_all_files.concat Dir[fdir]
		end
	end	
	
	#Putting the files in a Hash
	hsh_files = {}
	lst_all_files.each do |f|
		hsh_files[File.basename(f, File.extname(f))] = f
	end
	
	#Return the hash file
	hsh_files
end

#=================================================
# SCF License management
#=================================================

#LICENSE: Invoke the License dialog box or a message
def show_license
end

#LICENSE: Message for displaying the expiry date
def trial_expiry_compute
	@trial_expiry = Traductor.string_to_float_formula(@trial_expiry) if @trial_expiry.class == String
	@trial_expiry = 15 unless @trial_expiry
end

#LICENSE: Message for displaying the expiry date
def trial_expiry_message
	Traductor.use_libraries
	trial_expiry_compute
	tem_context(@trial_expiry, true) { |t| trial_expiry_message_display(t) }
end

def trial_expiry_message_display(t)
	trial_expiry_compute
	t = t + @trial_expiry * 3600 * 24 if t
	Traductor::TrialExpiryDialog.invoke(@plugin_name, t)
end

def trial_will_expire_notification(time_expiry)
	tc = Time.at(time_expiry)
	date = tc.strftime "%a %d %b %Y"
	nb_days = ((time_expiry - Time.now.to_f) / 86400.0).floor
	msg = "#{@plugin_name} will expire in #{nb_days} days on #{date}"
	key = "TrialExpire_#{nb_days}"
	if !Sketchup.read_default(@plugin_name, key)
		Sketchup.write_default(@plugin_name, key, true)
		G6.notification_OK(msg, :warning)
	end	
end

#=================================================
# About Dialog Box
#=================================================

#Show about dialog box
def show_about
	(SU_MAJOR_VERSION >= 6) ? show_about_v6 : show_about_v5
end
	
def show_about_v6
	AboutDialog.invoke self
end
	
def get_translation(symb, alter)
	s = T6[symb]
	(s == symb) ? s : Traductor[alter]
end
	
def html_about(html, div_id, main_div_height)
	#Initiate the fields
	name = get_translation :PLUGIN_Name, @lst_name
	description = get_translation :PLUGIN_Description, @lst_description
	written = (@written) ? @written : @creator
	created_by = T6[:T_HELP_CreatedBy, @creator] + ' - '
	copyright = (@copyright) ? created_by + T6[:T_HELP_Designed, written, Traductor[@lst_copyright]] : nil
	version = @version
	date = T6[@lst_date]
	title = "#{name} #{version} - #{date}"
	
	#style used in the dialog box
	space2 = "&nbsp;&nbsp;"
	skip = "<div style='height: 6px'>&nbsp</div>"
	
	html.create_style 'DivTable', nil, 'BD-SZ: 1', 'Bd: solid', 'Bd-col: gray', 'cellspacing: 0', 'align: center'
	html.create_style 'Title', nil, 'B', 'K: navy', 'F-SZ: 16'
	html.create_style 'Version', nil, 'K: green', 'F-SZ: 16'
	html.create_style 'Date', nil, 'K: black', 'F-SZ: 14'
	html.create_style 'Description', nil, 'B', 'K: blue', 'F-SZ: 12'
	html.create_style 'Label', nil, 'B', 'K: green', 'F-SZ: 11'
	html.create_style 'Info', nil, 'B', 'K: black', 'F-SZ: 11'
	html.create_style 'InfoS', nil, 'B', 'K: black', 'F-SZ: 10'
	html.create_style 'InfoSG', nil, 'B', 'K: green', 'F-SZ: 10'
	html.create_style 'InfoR', nil, 'B', 'K: red', 'F-SZ: 11'
	html.create_style 'Link', nil, 'B', 'K: blue', 'F-SZ: 9'
	html.create_style 'Copyright', nil, 'I', 'K: black', 'F-SZ: 11'
	html.create_style 'Section', nil, 'B', 'K: slateblue', 'BG: lightgrey', 'F-SZ: 11'
	html.create_style 'SectionS', nil, 'B', 'K: slateblue', 'BG: lightgrey', 'F-SZ: 10'
	html.create_style 'Credit', nil, 'K: black', 'F-SZ: 10'
	html.create_style 'CreditB', nil, 'B', 'K: RoyalBlue', 'F-SZ: 10'
	
	#Styling for screen and printing
	text = "<style type='text/css' media='screen'>"
	text << ".MAIN_DIV_Style {position: relative; height: #{main_div_height}px; overflow-y: auto; overflow-x: hidden; }"
	text << "</style>"
	html.body_add text
	
	#Creating the title
	txplug = HTML.format_span name, "", "Title"
	txversion = HTML.format_span version, "", "Version"
	txdate = HTML.format_span date, "", "Date"
	html.body_add "<div cellspacing='0px' cellpadding='0px' align='center'>", "#{txplug}&nbsp;#{txversion} - #{txdate}", "</div>"
	html.body_add skip
		
	#Description
	desc = HTML.format_span Traductor[@lst_description], "", "Description" if @lst_description
	copyright = HTML.format_span copyright, "", "Copyright" if copyright
	html.body_add "<table width='100%' cellpadding='3px'>"
	html.body_add "<tr><td align='center'>#{desc}</td></tr>"
	html.body_add "<tr><td align='center'>#{copyright}</td></tr>"
	html.body_add "</table>"
	html.body_add skip
		
	#Main div
	html.body_add "<div id='#{div_id}' width='100%' class='MAIN_DIV_Style DivTable'>"
	html.body_add "<table width='100%' cellspacing='0px' cellpadding='0px'>"
	
	#Information
	linfo = []
	linfo.push [T6[:T_HELP_Subfolder], @folder, 'Info']
	linfo.push [T6[:T_HELP_RootPath], @su_plugin_dir, 'InfoS', true]
	linfo.push [T6[:T_HELP_ParamPath], LibFredo6.defparam_dir, 'InfoS', true]
	linfo.push [T6[:T_HELP_TempPath], LibFredo6.tmpdir, 'InfoS', true]
	linfo.push [T6[:T_HELP_LoadTime], get_load_time, 'Info']
	if @plugin_name != "LibFredo6"
		linfo.push [T6[:T_HELP_LibFredo6], Plugin.libfredo6_version, 'InfoR']
	else	
		linfo.push [T6[:T_HELP_TotalLoadTime], Plugin.get_total_load_time, 'Info']
	end	
		
	text = HTML.format_span T6[:T_HELP_Info], "", "Section"
	html.body_add "<table width='100%' cellpadding='0px'>"
	html.body_add "<tr><td align='left'>#{text}</td></tr>"
	linfo.each do |info|
		label = HTML.format_span info[0], "", "Label"
		if info[3]
			text = info[1].gsub("fbessis", "fredo6")
			value = HTML.format_urllink text, "", info[2], nil, T6[:T_TIP_ClickOpenFolder], info[1]
		else
			value = HTML.format_span info[1], "", info[2]
		end	
		html.body_add "<tr><td>#{label} = #{value}</td></tr>"
	end
	html.body_add "</table>"
	html.body_add skip

	#Support
	if @web_site_name || @web_support_link || @web_site_link
		linfo = []
		linfo.push [:T_HELP_WebSiteName, @web_site_name, @web_site_link] if @web_site_name
		if @web_support_link
			linfo.push [:T_HELP_MenuWebInfo, @web_support_link]
		elsif @web_site_link
			linfo.push [:T_HELP_WebSiteLink, @web_site_link]
		end	
		
		if linfo.length > 0
			text = HTML.format_span T6[:T_HELP_Support], "", "Section"
			html.body_add "<table width='100%' cellpadding='0px'>"
			html.body_add "<tr><td align='left'>#{text}</td></tr>"
			linfo.each do |info|
				label = HTML.format_span T6[info[0]], "", "Label"
				value = HTML.format_textlink info[1], info[0].to_s, 'Link', nil, info[2]
				html.body_add "<tr><td>#{label} = #{value}</td></tr>"
			end
			html.body_add "</table>"
			html.body_add skip
		end	
	end	
		
	#Credit Table	
	if @lst_credits.length > 0
		text = HTML.format_span T6[:T_HELP_Credits], "", "Section"
		html.body_add "<table width='100%' cellpadding='0px'>"
		html.body_add "<tr><td colspan='2' align='left'>#{text}</td></tr>"
		
		@lst_credits.each do |credit|
			if credit =~ /:/
				text = HTML.format_span($`, "", "CreditB") + HTML.format_span($& + $', "", "Credit")
			else	
				text = HTML.format_span credit, "", "Credit"
			end	
			html.body_add "<tr><td width='10px' align='right'>-</td>"
			html.body_add "<td align='left'>#{text}</td></tr>"
		end	
		html.body_add "</table>"
	end

	#Usage
	lst_usage = usage_list
	unless lst_usage.empty?
		ntot, time_usage = usage_total
		text = HTML.format_span T6[:T_HELP_PLUGIN_USE], "", "Section"
		text << HTML.format_span(" (#{time_usage})", "", "InfoS")
		html.body_add "<table cellpadding='0px'>"
		html.body_add "<tr><td colspan='3' align='left'>#{text}</td></tr>"	
		lst_usage.each do |usage|
			text_cmd = HTML.format_span usage[1], "", "InfoS"
			text_nb = HTML.format_span "#{usage[0]}", "", "InfoSG"
			html.body_add "<tr><td align='left'>-</td>"
			html.body_add "<td align='left'>#{text_cmd}</td>"
			html.body_add "<td align='right'>#{text_nb}</td>"
			html.body_add "</tr>"
		end	
		html.body_add "</table>"
	end
	
	#End of scrolling DIV
	html.body_add "</table></div>"
	
	#Creating the dialog box button		
	dot = "..."
	butdone = HTML.format_button T6[:T_BUTTON_Done], "ButtonDone", 'R_Button', nil
	butprint = HTML.format_button T6[:T_BUTTON_Print], "ButtonPrint", 'N_Button', nil
	butcheck = HTML.format_button T6[:T_BUTTON_CheckForUpdate] + dot, "ButtonCheck", 'N_Button', nil
	butperf = HTML.format_button T6[:T_HELP_MenuPerformances] + dot, "ButtonPerf", 'N_Button', nil, T6[:T_TIP_MenuPerformances]
	butpurge = HTML.format_button T6[:T_HELP_MenuPurgeObsolete] + dot, "ButtonPurge", 'N_Button', nil, T6[:T_TIP_MenuPurgeObsolete]
	butlicense = ""
	if @scf_license
		butlicense = space2 + HTML.format_button(T6[:T_HELP_License] + dot, "ButtonLicense", 'Y_Button', nil, T6[:T_TIP_MenuLicense])
	end	
	html.body_add "<table class='T_NOPRINT_Style' width='99%' cellpadding='6px'><tr>"
	html.body_add "<td width='15%' align='left'>", butprint, "</td>"
	html.body_add "<td width='70%' align='center'>", butcheck, space2, butperf, space2, butpurge, butlicense, "</td>"
	html.body_add "<td align='right'>", butdone, "</td>"
	html.body_add "</tr></table>"
end
	
#Old version for SU5 which does not support Web dialog	
def show_about_v5
	text = ""
	text << Traductor[@lst_name] + " v" + @version
	text << "\n" + Traductor[@lst_description] if @lst_description
	text << "\n\n" + T6[:T_HELP_Designed, @creator, Traductor[@lst_copyright]]
	text << "\n" + T6[:T_HELP_Date] + " : " + T6[@lst_date]
	
	if @lst_credits.length > 0
		text << "\n\n" + T6[:T_HELP_Credits]
		@lst_credits.each { |w| text << "\n   - " + w }
	end	
	
	text << "\n\n" + T6[:T_HELP_Info]
	text << "\n  - " + T6[:T_HELP_RootPath] + " = " + @su_plugin_dir	
	text << "\n  - " + T6[:T_HELP_Subfolder] + " = " + @folder	
	text << "\n  - " + T6[:T_HELP_LoadTime] + " = #{get_load_time}"
	ntot, time_usage = usage_total
	text << "\n  - " + "#{T6[:T_HELP_Usage]} (#{time_usage})" + " = #{ntot}" if ntot
	if @plugin_name != "LibFredo6"
		text << "\n  - " + T6[:T_HELP_LibFredo6] + " = #{LibFredo6.folder} (#{MYPLUGIN.get_load_time})"
	else	
		text << "\n  - " + T6[:T_HELP_TotalLoadTime] + " = #{Plugin.get_total_load_time}"
	end	

	if @web_site_name || @web_support_link || @web_site_link
		text << "\n\n" + T6[:T_HELP_Support]
		text << "\n  - " + T6[:T_HELP_WebSiteName] + " --> " + @web_site_name if @web_site_name
		if @web_support_link
			text << "\n  - " + T6[:T_HELP_MenuWebInfo] + " --> " + @web_support_link
		elsif @web_site_link
			text << "\n  - " + T6[:T_HELP_WebSiteLink] + " --> " + @web_site_link
		end	
	end	
	
	UI.messagebox text, ((RUN_ON_MAC) ? MB_MULTILINE : MB_OK)
end

#Get the support link
def get_support_link
	@web_support_link
end

#Open the web page for a link
def open_support_link(ssymb)
	return unless ssymb.class == String
	case ssymb.intern
	when :T_HELP_WebSiteName, :T_HELP_WebSiteLink
		link = @web_site_link
	when :T_HELP_MenuWebInfo
		link = @web_support_link
	else
		return
	end
	Sketchup.set_status_text "Opening #{link}...."
	Traductor.openURL link
	Sketchup.set_status_text ""
end
	
#--------------------------------------------------------------------------------------------------------------
# Manage Old files and directories
#--------------------------------------------------------------------------------------------------------------			 

#Check older files - This method is obsolete
def check_older_scripts

end

#--------------------------------------------------------------------------------------------------------------
# Manage Configuration of a Plugin
#--------------------------------------------------------------------------------------------------------------			 

unless defined?(Traductor_Plugin_Command)
	Traductor_Plugin_Command = Struct.new :symb, :proc, :menutext, :icon, :ttip, :valproc, :icon_path,
                                          :test_cond, :separator, :submenu, :state, :cmd, :symb_log, :menu_tip_proc
end

def declare_toolbar(toolbarname, list_symb_buttons)
	@toolbarname = toolbarname
	list_symb_buttons = [list_symb_buttons] if list_symb_buttons.class == String
	@list_symb_buttons = list_symb_buttons
end

def declare_topmenu(topmenusymb=nil, list_menuperso=nil, separator=nil)
	@topmenu = (topmenusymb) ? eval_T6(topmenusymb) : nil
	if list_menuperso
		list_menuperso = [:T_MenuTools_Fredo6Collection, "Tools", :T_MenuPlugins_Fredo6Collection, "Plugins"] unless list_menuperso	
		list_menuperso = [list_menuperso] unless list_menuperso.class == Array
		@list_menuperso = list_menuperso
	end	
	@separator = separator
end

#Compute a nicer symbol for logging from the command symbol
def symb_log_from_symb(symb)
	s = symb.to_s
	s = $1 if s =~ /\A[A-Z]{1,3}_(.+)_|_command\Z/i || s =~ /\Aname_(.+)\Z/i || s =~ /\A(.+)_\Z/i
	s = s.gsub(/[A-Z]([A-Z]+)/) { |a| a[0..0] + a[1..-1].downcase } 
	ls = s.split(/_/)
	ls.each { |a| a[0..0] = a[0..0].upcase }
	s = ls.join ''
	s.intern
end

#Declare a command (i.e. both menu and icon in toolbar)
def declare_command(symb, iconroot=nil, symb_log=nil, &proc)
	symbtext = symb.to_s + "Menu"
	menutext = eval_T6 symbtext.intern
	symbttip = symb.to_s + "Tooltip"
	ttip = eval_T6 symbttip.intern
	ttip = menutext if ttip == symbttip
	ttip = @plugin_name + " - #{menutext}" + ": " + ttip unless ttip =~ /\:/
	declare_command_long symb, menutext, ttip, iconroot, symb_log, &proc
end

def declare_command_long(symb, text_menu, text_tooltip, iconroot=nil, symb_log=nil, &proc)
	tpc = Traductor_Plugin_Command.new
	tpc.symb = symb
	tpc.proc = proc
	tpc.menutext = text_menu
	symbttip = symb.to_s + "Tooltip"
	tpc.ttip = text_tooltip
	tpc.ttip = tpc.menutext unless tpc.ttip
	tpc.ttip = @plugin_name + " - #{text_menu}" + ": " + tpc.ttip unless tpc.ttip =~ /\:/
	tpc.icon = iconroot
	tpc.symb_log = (symb_log) ? symb_log : symb_log_from_symb(symb)
	tpc.icon = eval_constant(symb.to_s + "Icon") unless iconroot
	tpc.state = nil
	tpc.valproc = false
	@list_commands.push tpc
	@list_symb_commands.push symb
	@hsh_tpc[symb] = tpc
end

def list_commands ; @list_commands ; end
def list_symb_commands ; @list_symb_commands ; end

def match_command_from_symb(symb)
	return nil unless symb
	cmd = @hsh_commands[symb]
	return cmd if cmd
	
	@list_commands.each do |tpc|
		next if tpc.class == String
		tpc_symb = tpc.symb
		next unless tpc_symb
		if tpc_symb =~ /\A[A-Z1-9]+_(.+)/
			ssymb2 = $1
			ssymb2 = ssymb2[0..-2] if ssymb2[-1..-1] == '_'
			return @hsh_commands[tpc_symb] if symb.to_s.downcase == ssymb2.downcase
		end	
	end
	nil
end

def get_tpc(symb)
	tpc = @hsh_tpc[symb]
	return tpc if tpc
	
	pat = Regexp.new(symb.to_s, Regexp::IGNORECASE)
	@hsh_tpc.each do |sb, tpc|
		return tpc if sb =~ pat
	end	
	nil
end

def get_tpc_icon

end

def all_commands ; @hsh_commands ; end

#Return the Sketchup UI command for a symbol, or the full hash table if symb is nil
def get_command(symb=nil)
	(symb) ? match_command_from_symb(symb) : @hsh_commands
end

#Modify the tooltip of an icon (we cannot change the menu text or icon with the API!!)
def change_button_tooltip(symb, tooltip)
	cmd = @hsh_commands[symb]
	return unless cmd
	tooltip = "" unless tooltip
	cmd.tooltip = tooltip
	cmd.status_bar_text = tooltip.gsub(/\n/, " - ")
end

#Change the state of a button (to be used with care)
def set_button_state(symb, state=nil)
	tpc = @hsh_tpc[symb]
	return nil unless tpc
	tpc.state = state
	if state && !tpc.valproc
		tpc.cmd.set_validation_proc { state_validation_proc tpc.symb }
		tpc.valproc = true
	elsif state == nil && tpc.valproc
		tpc.valproc = false
		tpc.cmd.set_validation_proc { MF_ENABLED | MF_UNCHECKED}
	end	
end

def state_validation_proc(symb)
	tpc = @hsh_tpc[symb]
	(tpc.state) ? tpc.state : (MF_ENABLED | MF_UNCHECKED)
end

#Declare a contextual menu, possibly dependent on a condition
def declare_context_handler(symb, test_condition, separator=nil, submenu=nil, symb_log=nil, menu_tip_proc=nil, &proc)
	txh = symb.to_s + "Handler"
	text_menu = eval_T6 txh.intern
	symbtext = (symb.to_s + "Menu").intern
	text_menu = eval_T6 symbtext if text_menu == txh
	symbttip = symb.to_s + "Tooltip"
	text_tooltip = eval_T6 symbttip.intern
	text_tooltip = text_menu if text_tooltip == symbttip
	declare_context_handler_long symb, text_menu, text_tooltip, test_condition, separator, submenu, menu_tip_proc, &proc
end

def declare_context_handler_long(symb, text_menu, text_tooltip, test_condition, separator=nil, submenu=nil, symb_log=nil, menu_tip_proc=nil, &proc)
	tpc = Traductor_Plugin_Command.new
	tpc.symb = symb
	tpc.proc = proc
	tpc.menutext = text_menu
	tpc.ttip = text_tooltip
	tpc.ttip = tpc.menutext unless tpc.ttip
	tpc.test_cond = test_condition
	tpc.submenu = (submenu) ? eval_T6(submenu) : nil
	tpc.separator = separator
	tpc.menu_tip_proc = menu_tip_proc
	tpc.symb_log = (symb_log) ? symb_log : symb_log_from_symb(symb)
	@list_handlers.push tpc
end

#Declare a separator for Menu, Toolbar or both
def declare_separator(mt='MT')
	return unless mt
	@list_commands.push "M" if mt =~ /M/i
	@list_commands.push "T" if mt =~ /T/i
end

def declare_menu_separator
	@list_commands.push "M"
end

def declare_handler_separator
	@list_handlers.push "M"
end

def declare_toolbar_separator
	@list_commands.push "T"
end

def default_icons_visible(list_symb=nil)
	if list_symb
		list_symb = [list_symb] unless list_symb.class == Array
		@default_icons_visible |= list_symb.collect { |symb| symb.to_s }
	else
		@default_icons_visible = nil
	end
end

def default_handlers_visible(list_symb)
	if list_symb
		list_symb = [list_symb] unless list_symb.class == Array
		@default_handlers_visible |= list_symb.collect { |symb| symb.to_s }
	else
		@default_handlers_visible = []
	end
end

#Method to process all initialization task: commands, default parameters
def go
	@scf_license = (@scf_license && @scf_license =~ /true/i) ? true : false
	
	#Declaring the default param for the configuration
	build_config_defparam
	
	#Loading the configuration file
	@defparam.load_file
	
	#Check if the plugin is licensed
	if !@scf_license
		hsh, = tem_cebon(121)
		@scf_license = (hsh) ? true : nil
		@trial_expiry = 30
	end	
	
	#Configuring the toolbar and menu
	build_config_commands
	build_config_handlers
end

#Build the default parameters for the configuration
def build_config_defparam
	dp = @defparam
	
	#Title section
	dp.separator :T_DEFAULT_SECTION_Plugin
	
	#Alternate directory for pictures
	dp.alternate_icon_dir :__IMAGE_Dir, "", picture_all_folders
	declare_picture_folders_symb (:__IMAGE_Dir)

	#Toolbar
	dp.declare :T_DEFAULT_ToolbarName, @toolbar, 'T'
	
	#Icon shown in toolbar
	klist = []
	lsymb = []
	@list_commands.each do |tpc|
		next if tpc == 'M' || tpc == 'T'
		klist.push [tpc.symb.to_s, tpc.menutext]
		lsymb.push tpc.symb.to_s
	end
	if klist.length > 0
		lsymb = [] unless @default_icons_visible
		lsymb = @default_icons_visible if @default_icons_visible && @default_icons_visible.length > 0
		dp.declare :T_DEFAULT_IconVisible, lsymb.join(';;'), "M", eval_T6(klist)
	end	
	
	#Top menu
	lmenu = []
	if @list_menuperso
		@list_menuperso.each do |mp|
			symb = CustomMenu.get_menu_symb mp
			sumenu = CustomMenu.get_menu_sumenu mp
			text = CustomMenu.get_menu_name mp
			text = "" unless text
			text += " (#{sumenu})" if sumenu != ""
			lmenu.push [symb.to_s, text] if symb && text
		end	
	end	
	if lmenu.length > 1
		@defparam.declare :T_DEFAULT_TopMenu, lmenu[0][0], "H:", lmenu
	end	

	#Contextual menu
	klist = []
	lsymb = []
	@list_handlers.each do |tpc|
		next if tpc == 'M' || tpc == 'T'
		klist.push [tpc.symb.to_s, tpc.menutext]
		lsymb.push tpc.symb.to_s
	end
	if klist.length > 0
		lsymb = @default_handlers_visible unless @default_handlers_visible.empty?
		dp.declare :T_DEFAULT_HandlerVisible, lsymb.join(';;'), "M", eval_T6(klist)
	end	
	
end

#Get the main menu for the plugin
def get_main_submenu
	(@cmdfamily) ? @cmdfamily.get_menu : nil
end

@@hsh_menu_fredo_collection = {}

def compute_topmenu(topmenu_symb)
	topmenu_symb = "T_MenuPlugins_Fredo6Collection" unless topmenu_symb
	sumenu_name = 'Plugins'
	submenu_name = nil
	if topmenu_symb =~ /Plugins/i
		sumenu_name = 'Plugins'
	elsif topmenu_symb =~ /Tools/i
		sumenu_name = 'Tools'
	end	
	if topmenu_symb =~ /Fredo6Collection/i
		submenu_name = 'Fredo6 Collection'
	end	
		
	#No submenu
	return UI.menu(sumenu_name) unless submenu_name
	
	#Creating the submenu if not already done
	unless @@hsh_menu_fredo_collection[sumenu_name]
		sumenu = UI.menu(sumenu_name)
		@@hsh_menu_fredo_collection[sumenu_name] = sumenu.add_submenu(submenu_name)
	end	
	@@hsh_menu_fredo_collection[sumenu_name]			
end

#Build the commands (menus and toolbar buttons)
def build_config_commands	
	#Top menu and toolbar
	sumenu = compute_topmenu(@defparam[:T_DEFAULT_TopMenu])
	
	@topmenu = @plugin_name unless @topmenu
	@toolbar = @defparam[:T_DEFAULT_ToolbarName]
	@toolbar = @plugin_name unless @toolbar
	list_dir_icons = picture_selected_folders @defparam[@picture_symb].split(';;')
	@cmdfamily = Traductor::CommandFamily.new list_dir_icons, sumenu, @topmenu, @toolbar, @separator
	
	#Filtering the commands - Take only chosen icons and avoid double separators
	ltpc = []
	msep = true
	tsep = true
	lshown = @defparam[:T_DEFAULT_IconVisible]
	@list_commands.each do |tpc|
		if (tpc == 'M')
			ltpc.push tpc unless msep
			msep = true
		elsif tpc == 'T'
			ltpc.push tpc unless tsep
			tsep = true
		else
			vshow = lshown.include?(tpc.symb.to_s) && tpc.icon
			ltpc.push [vshow, tpc]
			icon16, icon24 = @cmdfamily.get_icons(tpc.icon, @icon_conv, nil)
			tpc.icon_path = (icon24) ? icon24 : icon16
			tsep = false if vshow
			msep = false
		end	
	end	
	return if ltpc.length == 0
	ltpc.pop until ltpc.last != 'M' && ltpc.last != 'T'	#trailing separators		
	return if ltpc.length == 0
	
	#Processing the commands
	pending_tsep = false
	ltpc.each do |tpc|
		if tpc == 'M'
			@cmdfamily.add_menu_separator
		elsif tpc == 'T'
			pending_tsep = true
		else
			vshow, tpc = tpc
			menutext = tpc.menutext
			ttip = tpc.ttip
			ttip = menutext unless ttip
			#icon = (vshow) ? tpc.icon : nil 
			icon = tpc.icon
			if icon && pending_tsep
				@cmdfamily.add_toolbar_separator
				pending_tsep = false
			end	
			cmd = @cmdfamily.add_command(menutext, ttip, icon, @icon_conv, nil) do
				command_invoke tpc 
			end	
			@cmdfamily.declare_invisible(cmd) unless vshow
			@hsh_commands[tpc.symb] = cmd if cmd
			tpc.cmd = cmd
		end	
	end
	
	#Standard support menu
	@cmdfamily.add_menu_separator
	populate_support_menu @cmdfamily, nil, nil
	
	#showing the toolbar
	@cmdfamily.show_toolbar
end

#Execute a command, and other tasks (usage log)
def command_invoke(tpc)
	return unless tpc
	usage_use tpc
	tpc.proc.call
end

#Register the usage of a command of the plugin
def usage_use(tpc)
	Traductor.use_libraries
	symb = tpc.symb
	usage_get
	n = @hsh_usage[symb.to_s]
	n = 0 unless n
	@hsh_usage[symb.to_s] = n + 1
	Default.store @plugin_name, "Usage", usage_encode
	#scf_usage_log(tpc, n+1)
end

#Get the total number of usages
def usage_total
	usage_get unless @hsh_usage
	t = @hsh_usage['_time']
	return nil unless t
	ntot = 0
	@hsh_usage.each { |key, val| ntot += val if key != '_time' }
	return nil if ntot == 0
	tc = Time.at t
	stime = tc.strftime "%a %d %b %Y - %Hh%M"
	[ntot, stime]
end

#return the total time and 
def usage_list
	usage_get unless @hsh_usage
	lst = []
	@hsh_usage.each do |key, val| 
		tpc = @hsh_tpc[key.intern]
		lst.push [val, tpc.menutext] if tpc
	end	
	lst.sort! { |a, b| a[1] <=> b[1] }
end

#Encode the string for usage to be stored in the registry
def usage_encode
	ls = []
	@hsh_usage.each do |key, value|
		ls.push "'#{key}'=>#{value}"
	end
	"{" + ls.join(',') + "}"
end

#Get the full information on usage
def usage_get
	unless @hsh_usage
		@hsh_usage = {}
		s = Default.read @plugin_name, "Usage"
		begin
			eval "@hsh_usage = #{s}" if s
		rescue
			return {}
		end
		@hsh_usage = { '_time' => Time.now.to_f } unless @hsh_usage['_time']
	end		
	@hsh_usage
end

#Build the commands (menus and toolbar buttons)
def build_config_handlers
	return if @list_handlers.length == 0
	
	$t1 = 0 unless $t1
	UI.add_context_menu_handler do |cxmenu|
		t1 = Time.now
		menu = cxmenu
		
		#Reinitializing the context for new menu
		if menu != @handler_menu
			@handler_menu = menu
			@hash_handler_menu = {}
		end	
		hcond = {}
		
		#Loop on contextual menus declared
		lshown = @defparam[:T_DEFAULT_HandlerVisible]
		nosep = true
		yasep = false
		@list_handlers.each do |tpc|
			t4 = Time.now
			
			#Separator
			if tpc == 'M'
				next if nosep
				yasep = true
				next
			end
			
			#not selected
			next unless lshown.include?(tpc.symb.to_s)
			
			#Evaluating the condition if any
			cond = tpc.test_cond
			if cond
				case hcond[cond.to_s]
				when 0
					#puts "*********** #{@rootname} - #{tpc.menutext} - Time = #{Time.now-t4} s"
					next
				when nil
					t0 = Time.now
					$t1 += Time.now - t0
					res = hcond[cond.to_s] = ((cond.execute_test) ? 1 : 0)
					#puts "   --> #{@rootname} - #{tpc.menutext} - Time = #{Time.now-t0} s T1 = #{$t1} s"
					#puts "*********** #{@rootname} - #{tpc.menutext} - Res = #{res} - Time = #{Time.now-t4} s"
					next if res == 0
				else	
					#puts "*********** #{@rootname} - #{tpc.menutext} - Time = #{Time.now-t4} s"
				end	
			end
			
			#Possible submenu
			t3 = Time.now
			submenu_text = tpc.submenu
			if submenu_text && submenu_text.strip != ''
				submenu = @hash_handler_menu[submenu_text]
				unless submenu
					menu.add_separator if tpc.separator
					submenu = menu.add_submenu submenu_text
					@hash_handler_menu[submenu_text] = submenu
					nosep = true
				end
				menu = submenu
			else
				menu = cxmenu
			end

			#Creating the item in the contextual menu
			menu.add_separator if (tpc.separator || yasep) && !nosep
			
			#Building the menu
			menutext = ttip = nil
			menu_tip_proc = tpc.menu_tip_proc
			if menu_tip_proc && menu_tip_proc.class == Proc
				menutext, ttip = menu_tip_proc.call
			end
			menutext = tpc.menutext unless menutext
			ttip = tpc.ttip unless ttip
			ttip = menutext unless ttip
			
			#Adding the menu
			menu.add_item(menutext) { command_invoke(tpc) }
			nosep = yasep = false
			#puts "*********** #{@rootname} - #{menutext} - Time = #{Time.now-t3} s"
		end	
		#puts "   --> #{@rootname} - Time = #{Time.now-t1} s"
	end 
end	

#Create a Cursor with the preferred image folders specified by Default Parameters
def create_cursor(name, hotx=0, hoty=0)
	unless @cursorfamily
		list_dir_icons = picture_selected_folders @defparam[@picture_symb].split(';;')
		@cursorfamily = Traductor::CursorFamily.new list_dir_icons, @cursor_conv
	end
	@cursorfamily.create_cursor name, hotx, hoty
end

#Compute a tooltip text with plugin name, version, date and author
def compute_tooltip
	"#{@plugin_name} v#{@version} - #{T6[@lst_date]} - #{@creator}"
end

#--------------------------------------------------------------------------------------------------------------
# TEM: Tempo management
#--------------------------------------------------------------------------------------------------------------

def tem_expired(t, hacked=false)
	Sketchup.active_model.select_tool nil
	#puts "TEM EXPIRED #{t.inspect} - time = #{Time.at(t)} - hacked = #{hacked}"
	trial_expiry_message_display(t)
end

def tem_encode_number(s)
	alphabet = {}
	alphabet['0'] = ['1', 'C', 'E', 'Z']
	alphabet['1'] = ['6', 'D', 'X', 'T']
	alphabet['2'] = ['3', 'K', 'F', 'P']
	alphabet['3'] = ['9', 'Q', 'U', 'S']
	alphabet['4'] = ['8', 'R', 'Y', 'M']
	alphabet['5'] = ['5', 'N', 'J']
	alphabet['6'] = ['7', 'V', 'L', 'O']
	alphabet['7'] = ['0', 'G', 'W']
	alphabet['8'] = ['4', 'A', 'H']
	alphabet['9'] = ['2', 'B', 'I']
		
	#Initialization
	e = ''
	action = s[0..0].to_i.modulo(2)
	s = s[1..-1]
		
	#Encoding
	if action == 0
		for i in 0..s.length-1
			j = s[i..i]
			al = alphabet[j]
			e += al[rand(al.length-1)]
		end
	
	#Decoding
	else
		for i in 0..s.length-1
			c = s[i..i]
			alphabet.each do |key, val|
				e += key if val.include?(c)
			end
		end
	end	
	
	#Returning the key
	e
end

#TEM: Check if a plugin is subject to TEM
def tem_cebon(key)
	stamp = Math.sqrt(key**2.5 + 7)
	case @rootname
	when 'JointPushPull', 'RoundCorner', 'FredoScale', 'ToolsOnSurface', 'Curviloft', 'VisuHole', 'TopoShaper', 'Curvizard'
		hsh = { :asof => Time.local(2022, 3, 1), :trial => 30 }
	#when 'ThruPaint'
	#	hsh = { :asof => Time.local(2023, 2, 1), :trial => 30 }
	else
		return [nil, stamp]
	end	
	
	hsh = nil if (hsh[:asof] - Time.now) >= 0
	[hsh, stamp]	
end

#TEM: Check whether the plugin is authorized to start	
def tem_launcher(ckey)
	cstamp = Math.sqrt(ckey**1.5 + 2)
	
	#Checking if the plugin is subject to TEM
	key = rand(100)
	hsh, stamp = tem_cebon(key)
	#puts "HSH = #{hsh} key = #{key} stamp = #{stamp} - equal = #{stamp == Math.sqrt(key**2.5 + 7)}"
	
	#OK to start the plugin
	return cstamp
	
	#Checking the trial period or license
	@xstamp = nil
	trial = hsh[:trial]
	tem_context(trial) { |xstamp| @xstamp = xstamp }
	
	#Return the stamp
	(@xstamp) ? cstamp : nil
end

#TEM: Tem facho check
def old_tem_facho(stamp)
	return false unless Sketchup.version.to_i == 17
	lst = Sketchup.find_support_file('Tools').split('/')
	return false if Sketchup.is_pro?
	return false unless lst.find { |a| a == 'SketchUp 2017' }
	(stamp / 27.0)**0.456 - (stamp / 17.0)**0.457
end

#TEM: TEM Management	
def tem_context(duration=nil, just_date=false, &tem_finish_proc)
	#*********************
	# CHECK LICENSE
	#*********************
	status = :ok
	return
end

#--------------------------------------------------------------------------------------------------------------
# Manage obsolete files
#--------------------------------------------------------------------------------------------------------------

#Register old files or directories
def register_obsolete_files(filedir)
	filedir = [filedir] unless filedir.class == Array
	@lst_obsolete_files = [] unless @lst_obsolete_files
	@lst_obsolete_files |= filedir
end

#Check the existence of obsolete files and return the list
def check_obsolete_files
	lfiles = []

	#Registered files
	$:.each do |sudir|
		@lst_obsolete_files.each do |f|
			lfiles += Dir[File.join(sudir, f)]
		end	
	end
	
	#Old directory of previous versions
	pat_old = Regexp.new("#{@plugin_name}_Dir_\\d\\d\\Z", Regexp::IGNORECASE)
	pat_new = Regexp.new("#{@plugin_name}_\\d\\d\\Z", Regexp::IGNORECASE)
	
	pdir = @plugin_dir + '/'
	$:.each do |sudir|
		Dir[File.join(sudir, "*#{@plugin_name}*")].each do |d|
			next if d == pdir && !File.directory?(d)
			b = File.basename d
			if b =~ pat_old || b =~ pat_new
				lfiles.push d
			end
		end	
	end
	lfiles
end

#--------------------------------------------------------------------------------------------------------------
# SCF: Manage statistics at SCF
#--------------------------------------------------------------------------------------------------------------

#SCF: store the usage of command and plugin for statistics
def scf_usage_log(tpc, n)
	return unless defined?(SCFapi) && defined?(SCFapi.storekeyvalue)
	
	#New plugins not registered yet in SCF
	return unless @web_support_link
	
	#Getting info on the command
	symb = tpc.symb
	symb_log = tpc.symb_log
	symb_log = symb unless symb_log
	key = @plugin_name + '-' + symb_log.to_s
	
	#Computing the threshold
	thresholds = [500, 100, 25, 10, 1]
	th = thresholds.find { |i| n >= i }

	#Getting the usage information
	dico = "LibFredo6_SCF"
	section = "UsageFunction"
	sparam = Default.read dico, section
	hsh = {}
	begin
		hsh = eval(sparam) if sparam
	rescue
		hsh = {}
	end

	#Checking if event is new
	val0 = "#{th}"
	val = hsh[key]
	return if val == val0
	
	#Updating the information
	hsh[key] = val0
	sparam = hsh.inspect.gsub('"', "'")
	Default.store dico, section, sparam	
	
	#Updating the SCF statistics
	author = "FrwGx547CiKm9qploXCD"
	key = "Usage"
	value = "#{plugin_name} - #{symb_log} - #{val0}"
	
	status = nil
	begin
		###status = SCFapi.storekeyvalue author, @plugin_name, key, value
	rescue
	end	
end

end	#class Plugin

#--------------------------------------------------------------------------------------------------------------
#--------------------------------------------------------------------------------------------------------------
# Custom Menu:
#--------------------------------------------------------------------------------------------------------------			 
#--------------------------------------------------------------------------------------------------------------			 

unless defined?(ALL_SU_MENUS)
	ALL_SU_MENUS = ["Plugins", "Tools", "File", "Edit", "View", "Camera", "Draw", "Page", "Window", "Help"]
end

class CustomMenu

unless defined?(Traductor_Custom_Menu)
	Traductor_Custom_Menu = Struct.new "Traductor_Custom_Menu", :symb, :menutext, :sumenu, :hmenu, :separator, :list_hmenus 
end

@@hsh_custom_menus = {}
@@hsh_custom_hmenus = {}

def CustomMenu.register(symb, menutext, sumenu, separator=false)
	hcm = @@hsh_custom_menus[symb.to_s]
	return hcm if hcm
	
	hcm = Traductor_Custom_Menu.new
	hcm.symb = symb
	hcm.menutext = (menutext) ? menutext : symb.to_s
	hcm.separator = separator
	hcm.sumenu = sumenu
	@@hsh_custom_menus[symb.to_s] = hcm
	hcm
end

#Get the SU handle to a custom menu
def CustomMenu.get_menu_suhandle(hcm)
	# SU top menu
	return hcm if ALL_SU_MENUS.include?(hcm)
	#return UI.menu(hcm) if ALL_SU_MENUS.include?(hcm)
	
	#Custom menu
	hcm = @@hsh_custom_menus[hcm.to_s] if hcm.class == String || hcm.class == Symbol
	return nil unless hcm.class == Struct::Traductor_Custom_Menu
	return hcm.menutext if hcm.hmenu
	
	#Creating the menu
	hmenu = get_menu_suhandle hcm.sumenu
	return nil unless hmenu
	hmenu.add_separator if hcm.separator
	hcm.hmenu = hmenu.add_submenu hcm.menutext
	@@hsh_custom_hmenus[hcm.hmenu.to_s] = hcm
	#return hcm.hmenu
	return hcm.menutext
end

def CustomMenu.get_menu_symb(hcm)
	return nil unless hcm
	return hcm if ALL_SU_MENUS.include?(hcm)
	hcm = @@hsh_custom_menus[hcm.to_s] if hcm.class == String || hcm.class == Symbol
	return nil unless hcm.class == Struct::Traductor_Custom_Menu
	hcm.symb
end

def CustomMenu.get_menu_name(hcm)
	return nil unless hcm
	return "Sketchup - #{hcm}" if ALL_SU_MENUS.include?(hcm)
	hcm = @@hsh_custom_menus[hcm.to_s] if hcm.class == String || hcm.class == Symbol
	return nil unless hcm.class == Struct::Traductor_Custom_Menu
	hcm.menutext
end

def CustomMenu.get_menu_sumenu(hcm)
	return nil unless hcm
	return "" if ALL_SU_MENUS.include?(hcm)
	hcm = @@hsh_custom_menus[hcm.to_s] if hcm.class == String || hcm.class == Symbol
	return nil unless hcm.class == Struct::Traductor_Custom_Menu
	CustomMenu.get_menu_name hcm.sumenu
end

end	#class Custom Menu

#--------------------------------------------------------------------------------------------------------------
#--------------------------------------------------------------------------------------------------------------
# Custom TestCondition: code attached to a test
#--------------------------------------------------------------------------------------------------------------			 
#--------------------------------------------------------------------------------------------------------------			 

class TestCondition

@@hsh_symb = {}

def initialize(symb=nil, &proc)
	@@hsh_symb[symb.to_s] = self
	@proc = proc	
	@val = nil
end

def execute_test
	@val = @proc.call if @proc
	@val
end

def get_val
	@val
end

def TestCondition.get_self(symb)
	return nil unless symb
	(symb.class == Traductor::TestCondition) ? symb : @@hsh_symb[symb.to_s]
end

end	#class TestCondition
	
end #Module Traductor

